@@ -49,6 +49,7 @@ from typing import (
     List,
     NoReturn,
     Optional,
+    Union,
 )
 
 
@@ -59,7 +60,7 @@ def displayaddress_handler(args: argparse.Namespace, client: HardwareWalletClien
     return displayaddress(client, desc=args.desc, path=args.path, addr_type=args.addr_type)
 
 def enumerate_handler(args: argparse.Namespace) -> List[Dict[str, Any]]:
-    return enumerate(password=args.password)
+    return enumerate(password=args.password, expert=args.expert, chain=args.chain)
 
 def getmasterxpub_handler(args: argparse.Namespace, client: HardwareWalletClient) -> Dict[str, str]:
     return getmasterxpub(client, addrtype=args.addr_type, account=args.account)
@@ -86,7 +87,7 @@ def setup_device_handler(args: argparse.Namespace, client: HardwareWalletClient)
 def signmessage_handler(args: argparse.Namespace, client: HardwareWalletClient) -> Dict[str, str]:
     return signmessage(client, message=args.message, path=args.path)
 
-def signtx_handler(args: argparse.Namespace, client: HardwareWalletClient) -> Dict[str, str]:
+def signtx_handler(args: argparse.Namespace, client: HardwareWalletClient) -> Dict[str, Union[bool, str]]:
     return signtx(client, psbt=args.psbt)
 
 def wipe_device_handler(args: argparse.Namespace, client: HardwareWalletClient) -> Dict[str, bool]:
@@ -135,7 +136,7 @@ def get_parser() -> HWIArgumentParser:
     parser = HWIArgumentParser(description='Hardware Wallet Interface, version {}.\nAccess and send commands to a hardware wallet device. Responses are in JSON format.'.format(__version__))
     parser.add_argument('--device-path', '-d', help='Specify the device path of the device to connect to')
     parser.add_argument('--device-type', '-t', help='Specify the type of device that will be connected. If `--device-path` not given, the first device of this type enumerated is used.')
-    parser.add_argument('--password', '-p', help='Device password if it has one (e.g. DigitalBitbox)', default='')
+    parser.add_argument('--password', '-p', help='Device password if it has one (e.g. DigitalBitbox)')
     parser.add_argument('--stdinpass', help='Enter the device password on the command line', action='store_true')
     parser.add_argument('--chain', help='Select chain to work with', type=Chain.argparse, choices=list(Chain), default=Chain.MAIN) # type: ignore
     parser.add_argument('--debug', help='Print debug statements', action='store_true')
@@ -276,12 +277,12 @@ def process_commands(cli_args: List[str]) -> Any:
 
     # Auto detect if we are using fingerprint or type to identify device
     if args.fingerprint or (args.device_type and not args.device_path):
-        client = find_device(args.password, args.device_type, args.fingerprint, args.expert)
+        client = find_device(args.password, args.device_type, args.fingerprint, args.expert, args.chain)
         if not client:
             return {'error': 'Could not find device with specified fingerprint', 'code': DEVICE_CONN_ERROR}
     elif args.device_type and args.device_path:
         with handle_errors(result=result, code=DEVICE_CONN_ERROR):
-            client = get_client(device_type, device_path, password, args.expert)
+            client = get_client(device_type, device_path, password, args.expert, args.chain)
         if 'error' in result:
             return result
     else:
@@ -290,8 +291,6 @@ def process_commands(cli_args: List[str]) -> Any:
     if client is None:
         return {"error": "Unable to communicated with device", "code": UNKNOWN_ERROR}
 
-    client.chain = args.chain
-
     # Do the commands
     with handle_errors(result=result, debug=args.debug):
         result = args.func(args, client)
