@@ -22,7 +22,7 @@ import importlib
 import logging
 import platform
 
-from ._base58 import xpub_to_pub_hex
+from ._base58 import xpub_to_pub_hex, xpub_to_xonly_pub_hex
 from .key import (
     get_bip44_purpose,
     get_bip44_chain,
@@ -42,6 +42,7 @@ from .descriptor import (
     Descriptor,
     parse_descriptor,
     MultisigDescriptor,
+    TRDescriptor,
     PKHDescriptor,
     PubkeyProvider,
     SHDescriptor,
@@ -51,6 +52,7 @@ from .descriptor import (
 from .devices import __all__ as all_devs
 from .common import (
     AddressType,
+    Chain,
 )
 from .hwwclient import HardwareWalletClient
 from .psbt import PSBT
@@ -61,6 +63,7 @@ from typing import (
     Dict,
     List,
     Optional,
+    Union,
 )
 
 
@@ -68,7 +71,7 @@ py_enumerate = enumerate
 
 
 # Get the client for the device
-def get_client(device_type: str, device_path: str, password: str = "", expert: bool = False) -> Optional[HardwareWalletClient]:
+def get_client(device_type: str, device_path: str, password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> Optional[HardwareWalletClient]:
     """
     Returns a HardwareWalletClient for the given device type at the device path
 
@@ -76,6 +79,7 @@ def get_client(device_type: str, device_path: str, password: str = "", expert: b
     :param device_path: The path specifying where the device can be accessed as returned by :func:`~enumerate`
     :param password: The password to use for this device
     :param expert: Whether the device should be opened in expert mode (prints more information for some commands)
+    :param chain: The Chain this client will be using
     :return: A :class:`~hwilib.hwwclient.HardwareWalletClient` to interact with the device
     :raises: UnknownDeviceError: if the device type is not known by HWI
     """
@@ -88,7 +92,7 @@ def get_client(device_type: str, device_path: str, password: str = "", expert: b
     try:
         imported_dev = importlib.import_module('.devices.' + module, __package__)
         client_constructor = getattr(imported_dev, class_name + 'Client')
-        client = client_constructor(device_path, password, expert)
+        client = client_constructor(device_path, password, expert, chain)
     except ImportError:
         if client:
             client.close()
@@ -97,7 +101,7 @@ def get_client(device_type: str, device_path: str, password: str = "", expert: b
     return client
 
 # Get a list of all available hardware wallets
-def enumerate(password: str = "") -> List[Dict[str, Any]]:
+def enumerate(password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> List[Dict[str, Any]]:
     """
     Enumerate all of the devices that HWI can potentially access.
 
@@ -110,7 +114,7 @@ def enumerate(password: str = "") -> List[Dict[str, Any]]:
     for module in all_devs:
         try:
             imported_dev = importlib.import_module('.devices.' + module, __package__)
-            result.extend(imported_dev.enumerate(password)) # type: ignore
+            result.extend(imported_dev.enumerate(password, expert, chain))
         except ImportError as e:
             # Warn for ImportErrors, but largely ignore them to allow users not install
             # all device dependencies if only one or some devices are wanted.
@@ -120,10 +124,11 @@ def enumerate(password: str = "") -> List[Dict[str, Any]]:
 
 # Fingerprint or device type required
 def find_device(
-    password: str = "",
+    password: Optional[str] = None,
     device_type: Optional[str] = None,
     fingerprint: Optional[str] = None,
     expert: bool = False,
+    chain: Chain = Chain.MAIN,
 ) -> Optional[HardwareWalletClient]:
     """
     Find a device from the device type or fingerprint and get a client to access it.
@@ -136,6 +141,7 @@ def find_device(
         The client returned will have a master public key fingerprint matching this.
         If not provided, device_type must be provided.
     :param expert: Whether the device should be opened in expert mode (enables additional output for some actions)
+    :param chain: The Chain this client will be using
     :return: A client to interact with the found device
     """
 
@@ -147,7 +153,7 @@ def find_device(
         try:
             assert isinstance(d["type"], str)
             assert isinstance(d["path"], str)
-            client = get_client(d['type'], d['path'], password, expert)
+            client = get_client(d['type'], d['path'], password, expert, chain)
             if client is None:
                 raise Exception()
 
@@ -176,7 +182,7 @@ def getmasterxpub(client: HardwareWalletClient, addrtype: AddressType = AddressT
     """
     return {"xpub": client.get_master_xpub(addrtype, account).to_string()}
 
-def signtx(client: HardwareWalletClient, psbt: str) -> Dict[str, str]:
+def signtx(client: HardwareWalletClient, psbt: str) -> Dict[str, Union[bool, str]]:
     """
     Sign a Partially Signed Bitcoin Transaction (PSBT) with the client.
 
@@ -188,7 +194,8 @@ def signtx(client: HardwareWalletClient, psbt: str) -> Dict[str, str]:
     # Deserialize the transaction
     tx = PSBT()
     tx.deserialize(psbt)
-    return {"psbt": client.sign_tx(tx).serialize()}
+    result = client.sign_tx(tx).serialize()
+    return {"psbt": result, "signed": result != psbt}
 
 def getxpub(client: HardwareWalletClient, path: str, expert: bool = False) -> Dict[str, Any]:
     """
@@ -287,8 +294,6 @@ def getdescriptor(
     :return: The descriptor constructed given the above arguments and key fetched from the device
     :raises: BadArgumentError: if an argument is malformed or missing.
     """
-    is_wpkh = addr_type is AddressType.WIT
-    is_sh_wpkh = addr_type is AddressType.SH_WIT
 
     parsed_path = []
     if not path:
@@ -334,12 +339,18 @@ def getdescriptor(
         client.xpub_cache[path_base] = client.get_pubkey_at_path(path_base).to_string()
 
     pubkey = PubkeyProvider(origin, client.xpub_cache.get(path_base, ""), path_suffix)
-    if is_wpkh:
-        return WPKHDescriptor(pubkey)
-    elif is_sh_wpkh:
-        return SHDescriptor(WPKHDescriptor(pubkey))
-    else:
+    if addr_type is AddressType.LEGACY:
         return PKHDescriptor(pubkey)
+    elif addr_type is AddressType.SH_WIT:
+        return SHDescriptor(WPKHDescriptor(pubkey))
+    elif addr_type is AddressType.WIT:
+        return WPKHDescriptor(pubkey)
+    elif addr_type is AddressType.TAP:
+        if not client.can_sign_taproot():
+            raise UnavailableActionError("Device does not support Taproot")
+        return TRDescriptor(pubkey)
+    else:
+        raise ValueError("Unknown address type")
 
 def getkeypool(
     client: HardwareWalletClient,
@@ -363,16 +374,21 @@ def getkeypool(
     :param internal: Whether the dictionary should indicate that the descriptor should be for change addresses
     :param keypool: Whether the dictionary should indicate that the dsecriptor should be added to the Bitcoin Core keypool/addresspool
     :param account: The BIP 44 account to use if ``path`` is not specified
-    :param sh_wpkh: Whether to return a descriptor specifying p2sh-segwit addresses
-    :param wpkh: Whether to return a descriptor specifying native segwit addresses
+    :param addr_type: The address type
     :param addr_all: Whether to return a multiple descriptors for every address type
     :return: The dictionary containing the descriptor and all of the arguments for ``importmulti`` or ``importdescriptors``
     :raises: BadArgumentError: if an argument is malformed or missing.
     """
+    supports_taproot = client.can_sign_taproot()
 
     addr_types = [addr_type]
     if addr_all:
         addr_types = list(AddressType)
+    elif not supports_taproot and addr_type == AddressType.TAP:
+        raise UnavailableActionError("Device does not support Taproot")
+
+    if not supports_taproot and AddressType.TAP in addr_types:
+        del addr_types[addr_types.index(AddressType.TAP)]
 
     # When no specific path or internal-ness is specified, create standard types
     chains: List[Dict[str, Any]] = []
@@ -404,7 +420,7 @@ def getdescriptors(
 
     for internal in [False, True]:
         descriptors = []
-        for addr_type in (AddressType.LEGACY, AddressType.SH_WIT, AddressType.WIT):
+        for addr_type in list(AddressType):
             try:
                 desc = getdescriptor(client, master_fpr=master_fpr, internal=internal, addr_type=addr_type, account=account)
             except UnavailableActionError:
@@ -446,12 +462,12 @@ def displayaddress(
         is_sh = isinstance(descriptor, SHDescriptor)
         is_wsh = isinstance(descriptor, WSHDescriptor)
         if is_sh or is_wsh:
-            assert descriptor.subdescriptor
-            descriptor = descriptor.subdescriptor
+            assert len(descriptor.subdescriptors) == 1
+            descriptor = descriptor.subdescriptors[0]
             if isinstance(descriptor, WSHDescriptor):
                 is_wsh = True
-                assert descriptor.subdescriptor
-                descriptor = descriptor.subdescriptor
+                assert len(descriptor.subdescriptors) == 1
+                descriptor = descriptor.subdescriptors[0]
             if isinstance(descriptor, MultisigDescriptor):
                 if is_sh and is_wsh:
                     addr_type = AddressType.SH_WIT
@@ -459,19 +475,21 @@ def displayaddress(
                     addr_type = AddressType.WIT
                 return {"address": client.display_multisig_address(addr_type, descriptor)}
         is_wpkh = isinstance(descriptor, WPKHDescriptor)
-        if isinstance(descriptor, PKHDescriptor) or is_wpkh:
+        if isinstance(descriptor, PKHDescriptor) or is_wpkh or isinstance(descriptor, TRDescriptor):
             pubkey = descriptor.pubkeys[0]
             if pubkey.origin is None:
                 raise BadArgumentError(f"Descriptor missing origin info: {desc}")
             if pubkey.origin.fingerprint != client.get_master_fingerprint():
                 raise BadArgumentError(f"Descriptor fingerprint does not match device: {desc}")
             xpub = client.get_pubkey_at_path(pubkey.origin.get_derivation_path()).to_string()
-            if pubkey.pubkey != xpub and pubkey.pubkey != xpub_to_pub_hex(xpub):
+            if pubkey.pubkey != xpub and pubkey.pubkey != xpub_to_pub_hex(xpub) and pubkey.pubkey != xpub_to_xonly_pub_hex(xpub):
                 raise BadArgumentError(f"Key in descriptor does not match device: {desc}")
             if is_sh and is_wpkh:
                 addr_type = AddressType.SH_WIT
             elif not is_sh and is_wpkh:
                 addr_type = AddressType.WIT
+            elif isinstance(descriptor, TRDescriptor):
+                addr_type = AddressType.TAP
             return {"address": client.display_singlesig_address(pubkey.get_full_derivation_path(0), addr_type)}
     raise BadArgumentError("Missing both path and descriptor")
 
