@@ -16,6 +16,7 @@ from typing import (
     Sequence,
     TypeVar,
 )
+import base64
 import builtins
 import sys
 from functools import wraps
@@ -29,6 +30,7 @@ from .._script import (
     is_p2pkh,
     is_p2wpkh,
     is_p2wsh,
+    is_p2tr,
     parse_multisig,
 )
 from ..psbt import PSBT
@@ -78,7 +80,6 @@ from bitbox02.communication.bitbox_api_protocol import (
     BitBoxNoiseConfig,
 )
 
-
 class BitBox02Error(UnavailableActionError):
     def __init__(self, msg: str):
         """
@@ -90,6 +91,7 @@ class BitBox02Error(UnavailableActionError):
         )
         msg += "m/49'/0'/<account'> for p2wpkh-p2sh; "
         msg += "m/84'/0'/<account'> for p2wpkh; "
+        msg += "m/86'/0'/<account'> for p2tr; "
         msg += "m/48'/0'/<account'>/2' for p2wsh multisig; "
         msg += "m/48'/0'/<account'>/1' for p2wsh-p2sh multisig; "
         msg += "m/48'/0'/<account'>' for any supported multisig; "
@@ -172,7 +174,7 @@ def _xpubs_equal_ignoring_version(xpub1: bytes, xpub2: bytes) -> bool:
     return xpub1[4:] == xpub2[4:]
 
 
-def enumerate(password: str = "") -> List[Dict[str, object]]:
+def enumerate(password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> List[Dict[str, Any]]:
     """
     Enumerate all BitBox02 devices. Bootloaders excluded.
     """
@@ -257,15 +259,15 @@ def bitbox02_exception(f: T) -> T:
 
 # This class extends the HardwareWalletClient for BitBox02 specific things
 class Bitbox02Client(HardwareWalletClient):
-    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
+    def __init__(self, path: str, password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> None:
         """
         Initializes a new BitBox02 client instance.
         """
-        super().__init__(path, password=password, expert=expert)
-        if password:
+        if password is not None:
             raise BadArgumentError(
                 "The BitBox02 does not accept a passphrase from the host. Please enable the passphrase option and enter the passphrase on the device during unlock."
             )
+        super().__init__(path, password=password, expert=expert, chain=chain)
 
         hid_device = hid.device()
         hid_device.open_path(path.encode())
@@ -285,17 +287,19 @@ class Bitbox02Client(HardwareWalletClient):
             return self.bb02
 
         for device_info in devices.get_any_bitbox02s():
-            if device_info["path"].decode() == self.device_path:
-                bb02 = bitbox02.BitBox02(
-                    transport=self.transport,
-                    device_info=device_info,
-                    noise_config=self.noise_config,
-                )
-                try:
-                    bb02.check_min_version()
-                except FirmwareVersionOutdatedException as exc:
-                    sys.stderr.write("WARNING: {}\n".format(exc))
-                    raise
+            if device_info["path"].decode() != self.device_path:
+                continue
+
+            bb02 = bitbox02.BitBox02(
+                transport=self.transport,
+                device_info=device_info,
+                noise_config=self.noise_config,
+            )
+            try:
+                bb02.check_min_version()
+            except FirmwareVersionOutdatedException as exc:
+                sys.stderr.write("WARNING: {}\n".format(exc))
+                raise
             self.bb02 = bb02
             is_initialized = bb02.device_info()["initialized"]
             if expect_initialized is not None:
@@ -336,7 +340,7 @@ class Bitbox02Client(HardwareWalletClient):
             "The BitBox02 does not need a PIN sent from the host"
         )
 
-    def _get_coin(self) -> bitbox02.btc.BTCCoin:
+    def _get_coin(self) -> "bitbox02.btc.BTCCoin.V":
         if self.chain != Chain.MAIN:
             return bitbox02.btc.TBTC
         return bitbox02.btc.BTC
@@ -361,6 +365,7 @@ class Bitbox02Client(HardwareWalletClient):
 
         - `m/49'/0'/<account'>` for `p2wpkh-p2sh` (segwit wrapped in P2SH)
         - `m/84'/0'/<account'>` for `p2wpkh` (native segwit v0)
+        - `m/86'/0'/<account'>` for `p2tr` (native segwit v1)
         - `m/48'/0'/<account'>/2'` for p2wsh multisig (native segwit v0 multisig).
         - `m/48'/0'/<account'>/1'` for p2wsh-p2sh multisig (p2sh-wrapped segwit v0 multisig).
         - `m/48'/0'/<account'>` for p2wsh and p2wsh-p2sh multisig.
@@ -402,7 +407,7 @@ class Bitbox02Client(HardwareWalletClient):
         self,
         threshold: int,
         origin_infos: Mapping[bytes, KeyOriginInfo],
-        script_type: bitbox02.btc.BTCScriptConfig.Multisig.ScriptType,
+        script_type: "bitbox02.btc.BTCScriptConfig.Multisig.ScriptType.V",
     ) -> Tuple[bytes, bitbox02.btc.BTCScriptConfigWithKeypath]:
         """
         From a threshold, {xpub: KeyOriginInfo} mapping and multisig script type,
@@ -463,6 +468,10 @@ class Bitbox02Client(HardwareWalletClient):
             raise UnavailableActionError(
                 "The BitBox02 does not support legacy p2pkh addresses"
             )
+        elif addr_type == AddressType.TAP:
+            script_config = bitbox02.btc.BTCScriptConfig(
+                simple_type=bitbox02.btc.BTCScriptConfig.P2TR
+            )
         else:
             raise BadArgumentError("Unknown address type")
         address = self.init().btc_address(
@@ -523,7 +532,7 @@ class Bitbox02Client(HardwareWalletClient):
         """
         Sign a transaction with the BitBox02.
 
-        he BitBox02 allows mixing inputs of different script types (e.g. and `p2wpkh-p2sh` `p2wpkh`), as
+        The BitBox02 allows mixing inputs of different script types (e.g. and `p2wpkh-p2sh` `p2wpkh`), as
         long as the keypaths use the appropriate bip44 purpose field per input (e.g. `49'` and `84'`) and
         all account indexes are the same.
 
@@ -533,7 +542,9 @@ class Bitbox02Client(HardwareWalletClient):
             keypaths: Dict[bytes, KeyOriginInfo]
         ) -> Tuple[Optional[bytes], Optional[Sequence[int]]]:
             """
-            Keypaths is a map of pubkey to hd keypath, where the first element in the keypath is the master fingerprint. We attempt to find the key which belongs to the BitBox02 by matching the fingerprint, and then matching the pubkey.
+            Keypaths is a map of pubkey to hd keypath, where the first element in the keypath is the master
+            fingerprint. We attempt to find the key which belongs to the BitBox02 by matching the fingerprint,
+            and then matching the pubkey.
             Returns the pubkey and the keypath, without the fingerprint.
             """
             for pubkey, origin in keypaths.items():
@@ -591,6 +602,13 @@ class Bitbox02Client(HardwareWalletClient):
                     ),
                     keypath=_keypath_hardened_prefix(keypath),
                 )
+            if is_p2tr(output.scriptPubKey):
+                return bitbox02.btc.BTCScriptConfigWithKeypath(
+                    script_config=bitbox02.btc.BTCScriptConfig(
+                        simple_type=bitbox02.btc.BTCScriptConfig.P2TR
+                    ),
+                    keypath=_keypath_hardened_prefix(keypath),
+                )
             # Check for segwit multisig (p2wsh or p2wsh-p2sh).
             is_p2wsh_p2sh = output.is_p2sh() and is_p2wsh(redeem_script)
             if output.is_p2wsh() or is_p2wsh_p2sh:
@@ -622,9 +640,11 @@ class Bitbox02Client(HardwareWalletClient):
         # must be exactly one pubkey per input that belongs to the BitBox02.
         found_pubkeys: List[bytes] = []
 
-        for input_index, (psbt_in, tx_in) in builtins.enumerate(
-            zip(psbt.inputs, psbt.tx.vin)
-        ):
+        for input_index, psbt_in in builtins.enumerate(psbt.inputs):
+            assert psbt_in.prev_txid is not None
+            assert psbt_in.prev_out is not None
+            assert psbt_in.sequence is not None
+
             if psbt_in.sighash and psbt_in.sighash != 1:
                 raise BadArgumentError(
                     "The BitBox02 only supports SIGHASH_ALL. Found sighash: {}".format(
@@ -642,27 +662,37 @@ class Bitbox02Client(HardwareWalletClient):
             # - https://medium.com/shiftcrypto/bitbox-app-firmware-update-6-2020-c70f733a5330
             # - https://blog.trezor.io/details-of-firmware-updates-for-trezor-one-version-1-9-1-and-trezor-model-t-version-2-3-1-1eba8f60f2dd.
             # - https://github.com/zkSNACKs/WalletWasabi/pull/3822
-            # The BitBox02 for now requires the prevtx, at least until Taproot activates.
+            # The BitBox02 requires all prevtxs if not all of the inputs are taproot.
 
             if psbt_in.non_witness_utxo:
-                if tx_in.prevout.hash != psbt_in.non_witness_utxo.sha256:
+                assert psbt_in.non_witness_utxo.sha256 is not None
+                if psbt_in.prev_txid != ser_uint256(psbt_in.non_witness_utxo.sha256):
                     raise BadArgumentError(
                         "Input {} has a non_witness_utxo with the wrong hash".format(
                             input_index
                         )
                     )
-                utxo = psbt_in.non_witness_utxo.vout[tx_in.prevout.n]
+                assert psbt_in.prev_out is not None
+                utxo = psbt_in.non_witness_utxo.vout[psbt_in.prev_out]
                 prevtx = psbt_in.non_witness_utxo
             elif psbt_in.witness_utxo:
                 utxo = psbt_in.witness_utxo
             if utxo is None:
                 raise BadArgumentError("No utxo found for input {}".format(input_index))
-            if prevtx is None:
-                raise BadArgumentError(
-                    "Previous transaction missing for input {}".format(input_index)
-                )
 
-            found_pubkey, keypath = find_our_key(psbt_in.hd_keypaths)
+            key_origin_infos = psbt_in.hd_keypaths.copy()
+            if len(psbt_in.tap_internal_key) > 0:
+                # adding taproot keys to the keypaths to be checked
+                for pubkey, (leaf_hashes, key_origin_info) in psbt_in.tap_bip32_paths.items():
+                    if len(leaf_hashes) > 0:
+                        raise BadArgumentError(
+                            "The BitBox02 does not support Taproot script path spending. Found leaf hashes: {}"
+                            .format(leaf_hashes)
+                        )
+                    key_origin_infos[pubkey] = key_origin_info
+
+            found_pubkey, keypath = find_our_key(key_origin_infos)
+
             if not found_pubkey:
                 raise BadArgumentError("No key found for input {}".format(input_index))
             assert keypath is not None
@@ -682,13 +712,13 @@ class Bitbox02Client(HardwareWalletClient):
             )
             inputs.append(
                 {
-                    "prev_out_hash": ser_uint256(tx_in.prevout.hash),
-                    "prev_out_index": tx_in.prevout.n,
+                    "prev_out_hash": psbt_in.prev_txid,
+                    "prev_out_index": psbt_in.prev_out,
                     "prev_out_value": utxo.nValue,
-                    "sequence": tx_in.nSequence,
+                    "sequence": psbt_in.sequence,
                     "keypath": keypath,
                     "script_config_index": script_config_index,
-                    "prev_tx": {
+                    "prev_tx": None if prevtx is None else {
                         "version": prevtx.nVersion,
                         "locktime": prevtx.nLockTime,
                         "inputs": [
@@ -712,10 +742,23 @@ class Bitbox02Client(HardwareWalletClient):
             )
 
         outputs: List[bitbox02.BTCOutputType] = []
-        for output_index, (psbt_out, tx_out) in builtins.enumerate(
-            zip(psbt.outputs, psbt.tx.vout)
-        ):
-            _, keypath = find_our_key(psbt_out.hd_keypaths)
+
+        for output_index, psbt_out in builtins.enumerate(psbt.outputs):
+            tx_out = psbt_out.get_txout()
+
+            key_origin_infos = psbt_out.hd_keypaths.copy()
+            if len(psbt_out.tap_internal_key) > 0:
+                # adding taproot keys to the keypaths to be checked
+                for pubkey, (leaf_hashes, key_origin_info) in psbt_out.tap_bip32_paths.items():
+                    if len(leaf_hashes) > 0:
+                        raise BadArgumentError(
+                            "The BitBox02 does not support Taproot script path spending. Found leaf hashes: {}"
+                            .format(leaf_hashes)
+                        )
+                    key_origin_infos.update({pubkey: key_origin_info})
+
+            _, keypath = find_our_key(key_origin_infos)
+
             is_change = keypath and keypath[-2] == 1
             if is_change:
                 assert keypath is not None
@@ -734,16 +777,19 @@ class Bitbox02Client(HardwareWalletClient):
             else:
                 if tx_out.is_p2pkh():
                     output_type = bitbox02.btc.P2PKH
-                    output_hash = tx_out.scriptPubKey[3:23]
+                    output_payload = tx_out.scriptPubKey[3:23]
                 elif is_p2wpkh(tx_out.scriptPubKey):
                     output_type = bitbox02.btc.P2WPKH
-                    output_hash = tx_out.scriptPubKey[2:]
+                    output_payload = tx_out.scriptPubKey[2:]
                 elif tx_out.is_p2sh():
                     output_type = bitbox02.btc.P2SH
-                    output_hash = tx_out.scriptPubKey[2:22]
+                    output_payload = tx_out.scriptPubKey[2:22]
                 elif is_p2wsh(tx_out.scriptPubKey):
                     output_type = bitbox02.btc.P2WSH
-                    output_hash = tx_out.scriptPubKey[2:]
+                    output_payload = tx_out.scriptPubKey[2:]
+                elif is_p2tr(tx_out.scriptPubKey):
+                    output_type = bitbox02.btc.P2TR
+                    output_payload = tx_out.scriptPubKey[2:]
                 else:
                     raise BadArgumentError(
                         "Output type not recognized of output {}".format(output_index)
@@ -752,7 +798,7 @@ class Bitbox02Client(HardwareWalletClient):
                 outputs.append(
                     bitbox02.BTCOutputExternal(
                         output_type=output_type,
-                        output_hash=output_hash,
+                        output_payload=output_payload,
                         value=tx_out.nValue,
                     )
                 )
@@ -766,26 +812,54 @@ class Bitbox02Client(HardwareWalletClient):
                 script_configs[0].script_config, script_configs[0].keypath
             )
 
+        assert psbt.tx_version is not None
         sigs = self.init().btc_sign(
             self._get_coin(),
             script_configs,
             inputs=inputs,
             outputs=outputs,
-            locktime=psbt.tx.nLockTime,
-            version=psbt.tx.nVersion,
+            locktime=psbt.compute_lock_time(),
+            version=psbt.tx_version,
         )
 
         for (_, sig), pubkey, psbt_in in zip(sigs, found_pubkeys, psbt.inputs):
             r, s = sig[:32], sig[32:64]
-            # ser_sig_der() adds SIGHASH_ALL
-            psbt_in.partial_sigs[pubkey] = ser_sig_der(r, s)
+
+            if len(psbt_in.tap_internal_key) > 0:
+                # taproot keypath input
+                psbt_in.tap_key_sig = sig
+            else:
+                # ser_sig_der() adds SIGHASH_ALL
+                psbt_in.partial_sigs[pubkey] = ser_sig_der(r, s)
 
         return psbt
 
+    @bitbox02_exception
     def sign_message(
         self, message: Union[str, bytes], bip32_path: str
     ) -> str:
-        raise UnavailableActionError("The BitBox02 does not support 'signmessage'")
+        if isinstance(message, str):
+            message = message.encode("utf-8")
+        keypath = parse_path(bip32_path)
+        purpose = keypath[0]
+        simple_type = {
+            PURPOSE_P2WPKH: bitbox02.btc.BTCScriptConfig.P2WPKH,
+            PURPOSE_P2WPKH_P2SH: bitbox02.btc.BTCScriptConfig.P2WPKH_P2SH,
+        }.get(purpose)
+        if simple_type is None:
+            raise BitBox02Error(
+                "For message signing, the keypath bip44 purpose must be 84' or 49'"
+            )
+        _, _, sig65 = self.init().btc_sign_msg(
+            self._get_coin(),
+            bitbox02.btc.BTCScriptConfigWithKeypath(
+                script_config=bitbox02.btc.BTCScriptConfig(
+                    simple_type=simple_type,
+                ),
+                keypath=keypath,
+            ),
+            message)
+        return base64.b64encode(sig65).decode("ascii")
 
     @bitbox02_exception
     def toggle_passphrase(self) -> bool:
@@ -841,3 +915,11 @@ class Bitbox02Client(HardwareWalletClient):
 
         bb02.restore_from_mnemonic()
         return True
+
+    def can_sign_taproot(self) -> bool:
+        """
+        The BitBox02 does not support Taproot yet.
+
+        :returns: False, always
+        """
+        return False
