@@ -67,6 +67,7 @@ from binascii import b2a_hex
 from typing import (
     Any,
     Callable,
+    Optional,
 )
 
 CC_SIMULATOR_SOCK = '/tmp/ckcc-simulator.sock'
@@ -89,8 +90,8 @@ def coldcard_exception(f: Callable[..., Any]) -> Callable[..., Any]:
 # This class extends the HardwareWalletClient for ColdCard specific things
 class ColdcardClient(HardwareWalletClient):
 
-    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
-        super(ColdcardClient, self).__init__(path, password, expert)
+    def __init__(self, path: str, password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> None:
+        super(ColdcardClient, self).__init__(path, password, expert, chain)
         # Simulator hard coded pipe socket
         if path == CC_SIMULATOR_SOCK:
             self.device = ColdcardDevice(sn=path)
@@ -142,6 +143,7 @@ class ColdcardClient(HardwareWalletClient):
 
         for _ in range(passes):
             # Get psbt in hex and then make binary
+            tx.convert_to_v0()
             fd = io.BytesIO(base64.b64decode(tx.serialize()))
 
             # learn size (portable way)
@@ -241,6 +243,8 @@ class ColdcardClient(HardwareWalletClient):
             addr_fmt = AF_P2WPKH
         elif addr_type == AddressType.LEGACY:
             addr_fmt = AF_CLASSIC
+        elif addr_type == AddressType.TAP:
+            raise UnavailableActionError("Coldcard does not support displaying Taproot addresses yet")
         else:
             raise BadArgumentError("Unknown address type")
 
@@ -386,7 +390,16 @@ class ColdcardClient(HardwareWalletClient):
         """
         raise UnavailableActionError('The Coldcard does not support toggling passphrase from the host')
 
-def enumerate(password: str = "") -> List[Dict[str, Any]]:
+    def can_sign_taproot(self) -> bool:
+        """
+        The Coldard does not support Taproot yet.
+
+        :returns: False, always
+        """
+        return False
+
+
+def enumerate(password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> List[Dict[str, Any]]:
     results = []
     devices = hid.enumerate(COINKITE_VID, CKCC_PID)
     devices.append({'path': CC_SIMULATOR_SOCK.encode()})
@@ -396,6 +409,7 @@ def enumerate(password: str = "") -> List[Dict[str, Any]]:
         path = d['path'].decode()
         d_data['type'] = 'coldcard'
         d_data['model'] = 'coldcard'
+        d_data['label'] = None
         d_data['path'] = path
         d_data['needs_pin_sent'] = False
         d_data['needs_passphrase_sent'] = False
