@@ -12,6 +12,7 @@ from typing import (
     NoReturn,
     Optional,
     Sequence,
+    Set,
     Tuple,
     Union,
 )
@@ -25,13 +26,15 @@ from ..errors import (
     DeviceConnectionError,
     DEVICE_NOT_INITIALIZED,
     DeviceNotReadyError,
+    NoPasswordError,
     UnavailableActionError,
     common_err_msgs,
     handle_errors,
 )
-from .trezorlib.client import TrezorClient as Trezor
+from .trezorlib.client import TrezorClient as Trezor, PASSPHRASE_ON_DEVICE
 from .trezorlib.debuglink import TrezorClientDebugLink
 from .trezorlib.exceptions import Cancelled, TrezorFailure
+from .trezorlib.models import TrezorModel
 from .trezorlib.transport import (
     DEV_TREZOR1,
     TREZORS,
@@ -48,6 +51,7 @@ from .._base58 import (
     get_xpub_fingerprint,
     to_address,
 )
+from .. import _base58 as base58
 
 from ..key import (
     ExtendedKey,
@@ -59,7 +63,12 @@ from .._script import (
     is_p2wsh,
     is_witness,
 )
-from ..psbt import PSBT
+from ..psbt import (
+    PSBT,
+    PartiallySignedInput,
+    PartiallySignedOutput,
+    KeyOriginInfo,
+)
 from ..tx import (
     CTxOut,
 )
@@ -69,6 +78,7 @@ from .._serialize import (
 from ..common import (
     AddressType,
     Chain,
+    hash256,
 )
 from .. import _bech32 as bech32
 from mnemonic import Mnemonic
@@ -76,6 +86,7 @@ from usb1 import USBErrorNoDevice
 from types import MethodType
 
 import base64
+import builtins
 import getpass
 import logging
 import sys
@@ -91,9 +102,22 @@ Use the numeric keypad to describe number positions. The layout is:
 
 Device = Union[hid.HidTransport, webusb.WebUsbTransport, udp.UdpTransport]
 
+ECDSA_SCRIPT_TYPES = [
+    messages.InputScriptType.SPENDADDRESS,
+    messages.InputScriptType.SPENDMULTISIG,
+    messages.InputScriptType.SPENDWITNESS,
+    messages.InputScriptType.SPENDP2SHWITNESS,
+]
+SCHNORR_SCRIPT_TYPES = [
+    messages.InputScriptType.SPENDTAPROOT,
+]
+
 
 # Only handles up to 15 of 15
-def parse_multisig(script: bytes) -> Tuple[bool, Optional[messages.MultisigRedeemScriptType]]:
+def parse_multisig(script: bytes, tx_xpubs: Dict[bytes, KeyOriginInfo], psbt_scope: Union[PartiallySignedInput, PartiallySignedOutput]) -> Tuple[bool, Optional[messages.MultisigRedeemScriptType]]:
+    # at least OP_M pub OP_N OP_CHECKMULTISIG
+    if len(script) < 37:
+        return (False, None)
     # Get m
     m = script[0] - 80
     if m < 1 or m > 15:
@@ -122,6 +146,19 @@ def parse_multisig(script: bytes) -> Tuple[bool, Optional[messages.MultisigRedee
     if op_cms != 174:
         return (False, None)
 
+    # check if we know corresponding xpubs from global scope
+    for pub in pubkeys:
+        if pub.node.public_key in psbt_scope.hd_keypaths:
+            derivation = psbt_scope.hd_keypaths[pub.node.public_key]
+            for xpub in tx_xpubs:
+                hd = ExtendedKey.deserialize(base58.encode(xpub + hash256(xpub)[:4]))
+                origin = tx_xpubs[xpub]
+                # check fingerprint and derivation
+                if (origin.fingerprint == derivation.fingerprint) and (origin.path == derivation.path[:len(origin.path)]):
+                    # all good - populate node and break
+                    pub.address_n = list(derivation.path[len(origin.path):])
+                    pub.node = messages.HDNodeType(depth=hd.depth, fingerprint=int.from_bytes(hd.parent_fingerprint, 'big'), child_num=hd.child_num, chain_code=hd.chaincode, public_key=hd.pubkey)
+                    break
     # Build MultisigRedeemScriptType and return it
     multisig = messages.MultisigRedeemScriptType(m=m, signatures=[b''] * n, pubkeys=pubkeys)
     return (True, multisig)
@@ -142,7 +179,7 @@ def trezor_exception(f: Callable[..., Any]) -> Any:
 
 
 def interactive_get_pin(self: object, code: Optional[int] = None) -> str:
-    if code == messages.PinMatrixRequestType.Currrent:
+    if code == messages.PinMatrixRequestType.Current:
         desc = "current PIN"
     elif code == messages.PinMatrixRequestType.NewFirst:
         desc = "new PIN"
@@ -211,7 +248,9 @@ class PassphraseUI:
     def disallow_passphrase(self) -> None:
         self.return_passphrase = False
 
-    def get_passphrase(self, available_on_device: bool) -> str:
+    def get_passphrase(self, available_on_device: bool) -> object:
+        if available_on_device:
+            return PASSPHRASE_ON_DEVICE
         if self.return_passphrase:
             return self.passphrase
         raise ValueError('Passphrase from Host is not allowed for Trezor T')
@@ -219,12 +258,18 @@ class PassphraseUI:
 
 HID_IDS = {DEV_TREZOR1}
 WEBUSB_IDS = TREZORS.copy()
+SIMULATOR_PATH = "127.0.0.1:21324"
 
 
-def get_path_transport(path: str) -> Device:
-    devs = hid.HidTransport.enumerate(usb_ids=HID_IDS)
-    devs.extend(webusb.WebUsbTransport.enumerate(usb_ids=WEBUSB_IDS))
-    devs.extend(udp.UdpTransport.enumerate())
+def get_path_transport(
+    path: str,
+    hid_ids: Set[Tuple[int, int]],
+    webusb_ids: Set[Tuple[int, int]],
+    sim_path: str
+) -> Device:
+    devs = hid.HidTransport.enumerate(usb_ids=hid_ids)
+    devs.extend(webusb.WebUsbTransport.enumerate(usb_ids=webusb_ids))
+    devs.extend(udp.UdpTransport.enumerate(sim_path))
     for dev in devs:
         if path == dev.get_path():
             return dev
@@ -234,17 +279,29 @@ def get_path_transport(path: str) -> Device:
 # This class extends the HardwareWalletClient for Trezor specific things
 class TrezorClient(HardwareWalletClient):
 
-    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
-        super(TrezorClient, self).__init__(path, password, expert)
+    def __init__(
+        self,
+        path: str,
+        password: Optional[str] = None,
+        expert: bool = False,
+        chain: Chain = Chain.MAIN,
+        hid_ids: Set[Tuple[int, int]] = HID_IDS,
+        webusb_ids: Set[Tuple[int, int]] = WEBUSB_IDS,
+        sim_path: str = SIMULATOR_PATH,
+        model: Optional[TrezorModel] = None
+    ) -> None:
+        if password is None:
+            password = ""
+        super(TrezorClient, self).__init__(path, password, expert, chain)
         self.simulator = False
-        transport = get_path_transport(path)
+        transport = get_path_transport(path, hid_ids, webusb_ids, sim_path)
         if path.startswith('udp'):
             logging.debug('Simulator found, using DebugLink')
-            self.client = TrezorClientDebugLink(transport=transport)
+            self.client = TrezorClientDebugLink(transport=transport, model=model, _init_device=False)
             self.simulator = True
             self.client.use_passphrase(password)
         else:
-            self.client = Trezor(transport=transport, ui=PassphraseUI(password))
+            self.client = Trezor(transport=transport, ui=PassphraseUI(password), model=model, _init_device=False)
 
         # if it wasn't able to find a client, throw an error
         if not self.client:
@@ -272,6 +329,17 @@ class TrezorClient(HardwareWalletClient):
             self.client.ui.disallow_passphrase()
         if self.client.features.pin_protection and not self.client.features.unlocked:
             raise DeviceNotReadyError('{} is locked. Unlock by using \'promptpin\' and then \'sendpin\'.'.format(self.type))
+        if self.client.features.passphrase_protection and self.password is None:
+            raise NoPasswordError("Passphrase protection is enabled, passphrase must be provided")
+
+    def _supports_external(self) -> bool:
+        if self.client.features.model == "1" and self.client.version <= (1, 10, 5):
+            return True
+        if self.client.features.model == "T" and self.client.version <= (2, 4, 3):
+            return True
+        if self.client.features.model == "K1-14AM":
+            return True
+        return False
 
     @trezor_exception
     def get_pubkey_at_path(self, path: str) -> ExtendedKey:
@@ -294,6 +362,7 @@ class TrezorClient(HardwareWalletClient):
         - Multisig inputs are limited to at most n-of-15 multisigs. This is a firmware limitation.
         - Transactions with arbitrary input scripts (scriptPubKey, redeemScript, or witnessScript) and arbitrary output scripts cannot be signed. This is a firmware limitation.
         - Send-to-self transactions will result in no prompt for outputs as all outputs will be detected as change.
+        - Transactions containing Taproot inputs cannot have external inputs.
         """
         self._check_unlocked()
 
@@ -309,11 +378,15 @@ class TrezorClient(HardwareWalletClient):
             # Prepare inputs
             inputs = []
             to_ignore = [] # Note down which inputs whose signatures we're going to ignore
-            for input_num, (psbt_in, txin) in py_enumerate(list(zip(tx.inputs, tx.tx.vin))):
+            for input_num, psbt_in in builtins.enumerate(tx.inputs):
+                assert psbt_in.prev_txid is not None
+                assert psbt_in.prev_out is not None
+                assert psbt_in.sequence is not None
+
                 txinputtype = messages.TxInputType(
-                    prev_hash=ser_uint256(txin.prevout.hash)[::-1],
-                    prev_index=txin.prevout.n,
-                    sequence=txin.nSequence,
+                    prev_hash=psbt_in.prev_txid[::-1],
+                    prev_index=psbt_in.prev_out,
+                    sequence=psbt_in.sequence,
                 )
 
                 # Detrermine spend type
@@ -322,9 +395,9 @@ class TrezorClient(HardwareWalletClient):
                 if psbt_in.witness_utxo:
                     utxo = psbt_in.witness_utxo
                 if psbt_in.non_witness_utxo:
-                    if txin.prevout.hash != psbt_in.non_witness_utxo.sha256:
+                    if psbt_in.prev_txid != psbt_in.non_witness_utxo.hash:
                         raise BadArgumentError('Input {} has a non_witness_utxo with the wrong hash'.format(input_num))
-                    utxo = psbt_in.non_witness_utxo.vout[txin.prevout.n]
+                    utxo = psbt_in.non_witness_utxo.vout[psbt_in.prev_out]
                 if utxo is None:
                     continue
                 scriptcode = utxo.scriptPubKey
@@ -339,13 +412,16 @@ class TrezorClient(HardwareWalletClient):
                     p2sh = True
 
                 # Check segwit
-                is_wit, _, _ = is_witness(scriptcode)
+                is_wit, wit_ver, _ = is_witness(scriptcode)
 
                 if is_wit:
-                    if p2sh:
-                        txinputtype.script_type = messages.InputScriptType.SPENDP2SHWITNESS
-                    else:
-                        txinputtype.script_type = messages.InputScriptType.SPENDWITNESS
+                    if wit_ver == 0:
+                        if p2sh:
+                            txinputtype.script_type = messages.InputScriptType.SPENDP2SHWITNESS
+                        else:
+                            txinputtype.script_type = messages.InputScriptType.SPENDWITNESS
+                    elif wit_ver == 1:
+                        txinputtype.script_type = messages.InputScriptType.SPENDTAPROOT
                 else:
                     txinputtype.script_type = messages.InputScriptType.SPENDADDRESS
                 txinputtype.amount = utxo.nValue
@@ -367,7 +443,7 @@ class TrezorClient(HardwareWalletClient):
                     to_ignore.append(input_num)
 
                 # Check for multisig
-                is_ms, multisig = parse_multisig(scriptcode)
+                is_ms, multisig = parse_multisig(scriptcode, tx.xpub, psbt_in)
                 if is_ms:
                     # Add to txinputtype
                     txinputtype.multisig = multisig
@@ -376,14 +452,20 @@ class TrezorClient(HardwareWalletClient):
                             txinputtype.script_type = messages.InputScriptType.SPENDMULTISIG
                         else:
                             # Cannot sign bare multisig, ignore it
+                            if not self._supports_external():
+                                raise BadArgumentError("Cannot sign bare multisig")
                             ignore_input()
                             continue
                 elif not is_ms and not is_wit and not is_p2pkh(scriptcode):
                     # Cannot sign unknown spk, ignore it
+                    if not self._supports_external():
+                        raise BadArgumentError("Cannot sign unknown scripts")
                     ignore_input()
                     continue
                 elif not is_ms and is_wit and p2wsh:
                     # Cannot sign unknown witness script, ignore it
+                    if not self._supports_external():
+                        raise BadArgumentError("Cannot sign unknown witness versions")
                     ignore_input()
                     continue
 
@@ -391,16 +473,29 @@ class TrezorClient(HardwareWalletClient):
                 found = False # Whether we have found a key to sign with
                 found_in_sigs = False # Whether we have found one of our keys in the signatures
                 our_keys = 0
-                for key in psbt_in.hd_keypaths.keys():
-                    keypath = psbt_in.hd_keypaths[key]
-                    if keypath.fingerprint == master_fp:
-                        if key in psbt_in.partial_sigs: # This key already has a signature
-                            found_in_sigs = True
-                            continue
-                        if not found: # This key does not have a signature and we don't have a key to sign with yet
-                            txinputtype.address_n = keypath.path
+                path_last_ours = None # The path of the last key that is ours. We will use this if we need to ignore this input because it is already signed.
+                if txinputtype.script_type in ECDSA_SCRIPT_TYPES:
+                    for key in psbt_in.hd_keypaths.keys():
+                        keypath = psbt_in.hd_keypaths[key]
+                        if keypath.fingerprint == master_fp:
+                            path_last_ours = keypath.path
+                            if key in psbt_in.partial_sigs: # This key already has a signature
+                                found_in_sigs = True
+                                continue
+                            if not found: # This key does not have a signature and we don't have a key to sign with yet
+                                txinputtype.address_n = keypath.path
+                                found = True
+                            our_keys += 1
+                elif txinputtype.script_type in SCHNORR_SCRIPT_TYPES:
+                    found_in_sigs = len(psbt_in.tap_key_sig) > 0
+                    for key, (leaf_hashes, origin) in psbt_in.tap_bip32_paths.items():
+                        # TODO: Support script path signing
+                        if key == psbt_in.tap_internal_key and origin.fingerprint == master_fp:
+                            path_last_ours = origin.path
+                            txinputtype.address_n = origin.path
                             found = True
-                        our_keys += 1
+                            our_keys += 1
+                            break
 
                 # Determine if we need to do more passes to sign everything
                 if our_keys > passes:
@@ -408,11 +503,16 @@ class TrezorClient(HardwareWalletClient):
 
                 if not found and not found_in_sigs: # None of our keys were in hd_keypaths or in partial_sigs
                     # This input is not one of ours
+                    if not self._supports_external():
+                        raise BadArgumentError("Cannot sign external inputs")
                     ignore_input()
                     continue
-                elif not found and found_in_sigs: # All of our keys are in partial_sigs, ignore whatever signature is produced for this input
-                    ignore_input()
-                    continue
+                elif not found and found_in_sigs:
+                    # All of our keys are in partial_sigs, pick the first key that is ours, sign with it,
+                    # and ignore whatever signature is produced for this input
+                    assert path_last_ours is not None
+                    txinputtype.address_n = path_last_ours
+                    to_ignore.append(input_num)
 
                 # append to inputs
                 inputs.append(txinputtype)
@@ -429,10 +529,14 @@ class TrezorClient(HardwareWalletClient):
 
             # prepare outputs
             outputs = []
-            for i, out in py_enumerate(tx.tx.vout):
+            for psbt_out in tx.outputs:
+                out = psbt_out.get_txout()
                 txoutput = messages.TxOutputType(amount=out.nValue)
                 txoutput.script_type = messages.OutputScriptType.PAYTOADDRESS
-                if out.is_p2pkh():
+                wit, ver, prog = out.is_witness()
+                if wit:
+                    txoutput.address = bech32.encode(bech32_hrp, ver, prog)
+                elif out.is_p2pkh():
                     txoutput.address = to_address(out.scriptPubKey[3:23], p2pkh_version)
                 elif out.is_p2sh():
                     txoutput.address = to_address(out.scriptPubKey[2:22], p2sh_version)
@@ -440,17 +544,13 @@ class TrezorClient(HardwareWalletClient):
                     txoutput.script_type = messages.OutputScriptType.PAYTOOPRETURN
                     txoutput.op_return_data = out.scriptPubKey[2:]
                 else:
-                    wit, ver, prog = out.is_witness()
-                    if wit:
-                        txoutput.address = bech32.encode(bech32_hrp, ver, prog)
-                    else:
-                        raise BadArgumentError("Output is not an address")
+                    raise BadArgumentError("Output is not an address")
 
-                # Add the derivation path for change, but only if there is exactly one derivation path
-                psbt_out = tx.outputs[i]
-                if len(psbt_out.hd_keypaths) == 1:
-                    _, keypath = next(iter(psbt_out.hd_keypaths.items()))
-                    if keypath.fingerprint == master_fp:
+                # Add the derivation path for change
+                if not wit or (wit and ver == 0):
+                    for _, keypath in psbt_out.hd_keypaths.items():
+                        if keypath.fingerprint != master_fp:
+                            continue
                         wit, ver, prog = out.is_witness()
                         if out.is_p2pkh():
                             txoutput.address_n = keypath.path
@@ -461,10 +561,27 @@ class TrezorClient(HardwareWalletClient):
                             txoutput.address = None
                         elif out.is_p2sh() and psbt_out.redeem_script:
                             wit, ver, prog = CTxOut(0, psbt_out.redeem_script).is_witness()
-                            if wit and len(prog) == 20:
+                            if wit and len(prog) in [20, 32]:
                                 txoutput.script_type = messages.OutputScriptType.PAYTOP2SHWITNESS
                                 txoutput.address_n = keypath.path
                                 txoutput.address = None
+                elif wit and ver == 1:
+                    for key, (leaf_hashes, origin) in psbt_out.tap_bip32_paths.items():
+                        # TODO: Support script path change
+                        if key == psbt_out.tap_internal_key and origin.fingerprint == master_fp:
+                            txoutput.address_n = origin.path
+                            txoutput.script_type = messages.OutputScriptType.PAYTOTAPROOT
+                            txoutput.address = None
+                            break
+
+                # add multisig info
+                if psbt_out.witness_script or psbt_out.redeem_script:
+                    is_ms, multisig = parse_multisig(
+                        psbt_out.witness_script or psbt_out.redeem_script,
+                        tx.xpub, psbt_out)
+                    if is_ms:
+                        txoutput.multisig = multisig
+                        txoutput.script_type = messages.OutputScriptType.PAYTOMULTISIG
 
                 # append to outputs
                 outputs.append(txoutput)
@@ -494,19 +611,22 @@ class TrezorClient(HardwareWalletClient):
                             script_pubkey=vout.scriptPubKey,
                         )
                         t.bin_outputs.append(o)
-                    logging.debug(psbt_in.non_witness_utxo.hash)
+                    assert(psbt_in.non_witness_utxo.hash is not None)
+                    logging.debug(psbt_in.non_witness_utxo.hash.hex())
                     assert psbt_in.non_witness_utxo.sha256 is not None
                     prevtxs[ser_uint256(psbt_in.non_witness_utxo.sha256)[::-1]] = t
 
             # Sign the transaction
+            assert tx.tx_version is not None
             signed_tx = btc.sign_tx(
                 client=self.client,
                 coin_name=self.coin_name,
                 inputs=inputs,
                 outputs=outputs,
                 prev_txes=prevtxs,
-                version=tx.tx.nVersion,
-                lock_time=tx.tx.nLockTime,
+                version=tx.tx_version,
+                lock_time=tx.compute_lock_time(),
+                serialize=False,
             )
 
             # Each input has one signature
@@ -518,6 +638,10 @@ class TrezorClient(HardwareWalletClient):
                     if fp == master_fp and pubkey not in psbt_in.partial_sigs:
                         psbt_in.partial_sigs[pubkey] = sig + b'\x01'
                         break
+                if len(psbt_in.tap_internal_key) > 0 and len(psbt_in.tap_key_sig) == 0:
+                    # Assume key path sig
+                    # TODO: Deal with script path sig
+                    psbt_in.tap_key_sig = sig
 
             p += 1
 
@@ -545,6 +669,10 @@ class TrezorClient(HardwareWalletClient):
             script_type = messages.InputScriptType.SPENDWITNESS
         elif addr_type == AddressType.LEGACY:
             script_type = messages.InputScriptType.SPENDADDRESS
+        elif addr_type == AddressType.TAP:
+            if not self.can_sign_taproot():
+                raise UnavailableActionError("This device does not support displaying Taproot addresses")
+            script_type = messages.InputScriptType.SPENDTAPROOT
         else:
             raise BadArgumentError("Unknown address type")
 
@@ -629,7 +757,7 @@ class TrezorClient(HardwareWalletClient):
 
         if self.client.features.initialized:
             raise DeviceAlreadyInitError('Device is already initialized. Use wipe first and try again')
-        device.reset(self.client, passphrase_protection=bool(self.password))
+        device.reset(self.client, label=label or None, passphrase_protection=bool(self.password))
         return True
 
     @trezor_exception
@@ -645,7 +773,7 @@ class TrezorClient(HardwareWalletClient):
             # Use interactive_get_pin
             self.client.ui.get_pin = MethodType(interactive_get_pin, self.client.ui)
 
-        device.recover(self.client, word_count=word_count, label=label, input_callback=mnemonic_words(), passphrase_protection=bool(self.password))
+        device.recover(self.client, word_count=word_count, label=label or None, input_callback=mnemonic_words(), passphrase_protection=bool(self.password))
         return True
 
     def backup_device(self, label: str = "", passphrase: str = "") -> bool:
@@ -689,7 +817,7 @@ class TrezorClient(HardwareWalletClient):
                     raise DeviceAlreadyUnlockedError('The PIN has already been sent to this device')
             return False
         elif isinstance(resp, messages.PassphraseRequest):
-            pass_resp = self.client.call_raw(messages.PassphraseAck(passphrase=self.client.ui.get_passphrase(available_on_device=False), on_device=False))
+            pass_resp = self.client.call(messages.PassphraseAck(passphrase=self.client.ui.get_passphrase(available_on_device=False), on_device=False), check_fw=False)
             if isinstance(pass_resp, messages.Deprecated_PassphraseStateRequest):
                 self.client.call_raw(messages.Deprecated_PassphraseStateAck())
         return True
@@ -706,7 +834,21 @@ class TrezorClient(HardwareWalletClient):
                 print(PIN_MATRIX_DESCRIPTION, file=sys.stderr)
         return True
 
-def enumerate(password: str = "") -> List[Dict[str, Any]]:
+    @trezor_exception
+    def can_sign_taproot(self) -> bool:
+        """
+        Trezor T supports Taproot since firmware version 2.4.3.
+        Trezor One supports Taproot since firmware version 1.10.4.
+
+        :returns: False, always.
+        """
+        self._prepare_device()
+        if self.client.features.model == "T":
+            return bool(self.client.version >= (2, 4, 3))
+        return bool(self.client.version >= (1, 10, 4))
+
+
+def enumerate(password: Optional[str] = None, expert: bool = False, chain: Chain = Chain.MAIN) -> List[Dict[str, Any]]:
     results = []
     devs = hid.HidTransport.enumerate()
     devs.extend(webusb.WebUsbTransport.enumerate())
@@ -727,8 +869,9 @@ def enumerate(password: str = "") -> List[Dict[str, Any]]:
             if 'trezor' not in client.client.features.vendor:
                 continue
 
+            d_data['label'] = client.client.features.label
             d_data['model'] = 'trezor_' + client.client.features.model.lower()
-            if d_data['path'] == 'udp:127.0.0.1:21324':
+            if d_data['path'].startswith('udp:'):
                 d_data['model'] += '_simulator'
 
             d_data['needs_pin_sent'] = client.client.features.pin_protection and not client.client.features.unlocked
@@ -738,8 +881,8 @@ def enumerate(password: str = "") -> List[Dict[str, Any]]:
                 d_data['needs_passphrase_sent'] = False
             if d_data['needs_pin_sent']:
                 raise DeviceNotReadyError('Trezor is locked. Unlock by using \'promptpin\' and then \'sendpin\'.')
-            if d_data['needs_passphrase_sent'] and not password:
-                raise DeviceNotReadyError("Passphrase needs to be specified before the fingerprint information can be retrieved")
+            if d_data['needs_passphrase_sent'] and password is None:
+                d_data["warnings"] = [["Passphrase protection enabled but passphrase was not provided. Using default passphrase of the empty string (\"\")"]]
             if client.client.features.initialized:
                 d_data['fingerprint'] = client.get_master_fingerprint().hex()
                 d_data['needs_passphrase_sent'] = False # Passphrase is always needed for the above to have worked, so it's already sent
